@@ -6,13 +6,14 @@ from enum import Enum
 from typing import List, Union, NamedTuple
 import inspect
 from collections import OrderedDict
+from dataclasses import dataclass, field
 
 import minidump.minidumpfile as minidump
 from unicorn import *
 from unicorn.x86_const import *
 from pefile import *
 
-from .handles import HandleManager, FileObject
+from .handles import *
 from .native import *
 from .details import *
 from .memory import *
@@ -37,29 +38,29 @@ class ExceptionType(Enum):
     Interrupt = 2
     ContextSwitch = 3
 
+@dataclass
 class ExceptionInfo:
-    def __init__(self):
-        self.type = ExceptionType.NoException
-        self.memory_access = 0
-        self.memory_address = 0
-        self.memory_size = 0
-        self.memory_value = 0
-        self.interrupt_number = 0
-        self.code_hook_h: Optional[int] = None  # TODO: should be unicorn.uc_hook_h, but type error
-        self.context: Optional[unicorn.UcContext] = None
-        self.tb_start = 0
-        self.tb_size = 0
-        self.tb_icount = 0
-        self.step_count = 0
-        self.final = False
-        self.handling = False
+    type: ExceptionType = ExceptionType.NoException
+    memory_access: int = 0  # refers to UC_MEM_* values
+    memory_address: int = 0
+    memory_size: int = 0
+    memory_value: int = 0
+    interrupt_number: int = 0
+    code_hook_h: Optional[int] = None  # represents a `unicorn.uc_hook_h` value (from uc.hook_add)
+    context: Optional[unicorn.UcContext] = None
+    tb_start: int = 0
+    tb_size: int = 0
+    tb_icount: int = 0
+    step_count: int = 0
+    final: bool = False
+    handling: bool = False
 
     def __str__(self):
         return f"{self.type}, ({hex(self.tb_start)}, {hex(self.tb_size)}, {self.tb_icount})"
 
+@dataclass
 class UnicornPageManager(PageManager):
-    def __init__(self, uc: Uc) -> None:
-        self._uc = uc
+    _uc: Uc
 
     def commit(self, addr: int, size: int, protect: MemoryProtect) -> None:
         perms = map_unicorn_perms(protect)
@@ -79,30 +80,32 @@ class UnicornPageManager(PageManager):
             data = bytes(data)
         self._uc.mem_write(addr, data)
 
+@dataclass
 class LazyPage:
-    def __init__(self, addr: int, protect: MemoryProtect, committed: bool):
-        self.addr = addr
-        self.protect = protect
-        self.committed = committed
-        self.data: Optional[bytearray] = None
+    addr: int
+    protect: MemoryProtect
+    committed: bool
+    data: Optional[bytearray] = None
 
     @property
     def size(self):
         return PAGE_SIZE
 
+@dataclass
 class LazyPageManager(PageManager):
-    def __init__(self, child: PageManager):
-        self.child = child
-        self.total_commit = 0
-        self.pages: Dict[int, LazyPage] = {}
-        self.lazy = True
+    child: PageManager
+    total_commit: int = 0
+    pages: Dict[int, LazyPage] = field(default_factory=dict)
+    lazy: bool = True
 
-    def iter_pages(self, addr: int, size: int):
+    @staticmethod
+    def iter_pages(addr: int, size: int):
         for i in range(0, size // PAGE_SIZE):
             page_addr = addr + i * PAGE_SIZE
             yield page_addr
 
-    def iter_chunks(self, addr: int, size: int):
+    @staticmethod
+    def iter_chunks(addr: int, size: int):
         # TODO: rewrite this to not be so disgusting
         page = addr & ~0xFFF
         index = addr & 0xFFF
@@ -136,7 +139,7 @@ class LazyPageManager(PageManager):
     def commit(self, addr: int, size: int, protect: MemoryProtect) -> None:
         assert addr & 0xFFF == 0
         assert size & 0xFFF == 0
-        #print(f"commit({hex(addr)}, {hex(size)}, {protect})")
+
         if not self.lazy:
             self.child.commit(addr, size, protect)
         for page_addr in self.iter_pages(addr, size):
@@ -147,7 +150,7 @@ class LazyPageManager(PageManager):
     def decommit(self, addr: int, size: int) -> None:
         assert addr & 0xFFF == 0
         assert size & 0xFFF == 0
-        #print(f"decommit({hex(addr)}, {hex(size)})")
+
         pages = []
         for page_addr in self.iter_pages(addr, size):
             assert page_addr in self.pages
@@ -166,7 +169,7 @@ class LazyPageManager(PageManager):
     def protect(self, addr: int, size: int, protect: MemoryProtect) -> None:
         assert addr & 0xFFF == 0
         assert size & 0xFFF == 0
-        #print(f"protect({hex(addr)}, {hex(size)}, {protect})")
+
         pages = []
         for page_addr in self.iter_pages(addr, size):
             assert page_addr in self.pages
@@ -183,12 +186,11 @@ class LazyPageManager(PageManager):
             page.protect = protect
 
     def read(self, addr: int, size: int) -> bytearray:
-        #print(f"read({hex(addr)}, {hex(size)})")
-
         pages = []
         for page_addr, index, length in self.iter_chunks(addr, size):
             page = self.pages.get(page_addr, None)
-            assert page is not None
+            if page is None:
+                raise IndexError(f"Could not find page {hex(page_addr)} while reading {hex(addr)}[{hex(size)}]")
             pages.append((page, index, length))
 
         if all([page.committed for page, _, _ in pages]):
@@ -208,12 +210,11 @@ class LazyPageManager(PageManager):
             return data
 
     def write(self, addr: int, data: bytes) -> None:
-        #print(f"write({hex(addr)}, size: {hex(len(data))})")
-
         pages = []
         for page_addr, index, length in self.iter_chunks(addr, len(data)):
             page = self.pages.get(page_addr, None)
-            assert page is not None
+            if page is None:
+                raise IndexError(f"Could not find page {hex(page_addr)} while writing {hex(addr)}[{hex(len(data))}]")
             pages.append((page, index, length))
 
         if all([page.committed for page, _, _ in pages]):
@@ -233,7 +234,7 @@ class LazyPageManager(PageManager):
 
 class SimpleTimer:
     def __init__(self):
-        self.time = 0
+        self.time = 0.0
         self.start()
 
     def start(self):
@@ -250,6 +251,7 @@ class Dumpulator(Architecture):
     def __init__(self, minidump_file, *, trace=False, quiet=False, thread_id=None, debug_logs=False):
         self._quiet = quiet
         self._debug = debug_logs
+        self.sequence_id = 0
 
         # Load the minidump
         self._minidump = minidump.MinidumpFile.parse(minidump_file)
@@ -291,10 +293,12 @@ class Dumpulator(Architecture):
         self.syscalls = []
         self._setup_syscalls()
         self._setup_emulator(thread)
+        self.handles = HandleManager()
+        self._setup_handles()
+        self._setup_registry()
         self.kill_me = None
         self.exit_code = None
         self.exports = self._all_exports()
-        self.handles = HandleManager()
         self.exception = ExceptionInfo()
         self.last_exception: Optional[ExceptionInfo] = None
         if not self._quiet:
@@ -341,7 +345,8 @@ class Dumpulator(Architecture):
         if not self._quiet:
             print(message)
 
-    def error(self, message: str):
+    @staticmethod
+    def error(message: str):
         print(message)
 
     def _switch_segment(self, segment: SegmentRegisters, gs_base: Optional[int] = None, fs_base: Optional[int] = None):
@@ -390,7 +395,6 @@ class Dumpulator(Architecture):
                     reserve_addr = info.BaseAddress
                 reserve_size += info.RegionSize
             info = region[0]
-            emu_addr = info.BaseAddress & self.addr_mask
             if info.State == minidump.MemoryState.MEM_FREE:
                 continue
             reserve_protect = MemoryProtect(info.AllocationProtect)
@@ -462,6 +466,7 @@ class Dumpulator(Architecture):
             self.stdin_handle = self.read_ptr(process_parameters + 0x20)
             self.stdout_handle = self.read_ptr(process_parameters + 0x28)
             self.stderr_handle = self.read_ptr(process_parameters + 0x30)
+            self.modules.main = self.read_ptr(self.peb + 0x10)
             number_of_heaps = self.read_ulong(self.peb + 0xe8)
             process_heaps_ptr = self.read_ptr(self.peb + 0xf0)
             api_set_map = self.read_ptr(self.peb + 0x68)
@@ -482,6 +487,7 @@ class Dumpulator(Architecture):
             self.stdin_handle = self.read_ptr(process_parameters + 0x18)
             self.stdout_handle = self.read_ptr(process_parameters + 0x1c)
             self.stderr_handle = self.read_ptr(process_parameters + 0x20)
+            self.modules.main = self.read_ptr(self.peb + 0x8)
             number_of_heaps = self.read_ulong(self.peb + 0x88)
             process_heaps_ptr = self.read_ptr(self.peb + 0x90)
             api_set_map = self.read_ptr(self.peb + 0x38)
@@ -518,6 +524,80 @@ class Dumpulator(Architecture):
         self.memory.set_region_info(default_activation_context_data, "Default activation context data")
         self.memory.set_region_info(leap_second_data, "Leap second data")
 
+    def _setup_registry(self):
+        self.handles.create_key(r"\Registry\Machine\System\CurrentControlSet\Control\Nls\Sorting\Versions", {
+            "": "00060305",
+            "000601xx": "SortWindows61.dll",
+            "000602xx": "SortWindows62.dll",
+            "000603xx": "kernel32.dll",
+            "FF0000xx": "SortServer2003Compat.dll",
+            "FF0406xx": "SortWindows6Compat.dll",
+            "FF0502xx": "SortWindows6Compat.dll",
+            "000604xx": "SortWindows64.dll",
+        })
+
+    def _setup_handles(self):
+        import dumpulator.ntdevices as ntdevices
+        self.console = ntdevices.ConsoleDeviceObject(R"\Device\ConDrv")
+        self.stdin = ConsoleFileObject(ConsoleType.In)
+        self.stdout = ConsoleFileObject(ConsoleType.Out)
+        self.stderr = ConsoleFileObject(ConsoleType.Err)
+
+        if self.console_handle != 0:
+            self.handles.add(self.console_handle, self.console)
+        if self.stdin_handle != 0:
+            self.handles.add(self.stdin_handle, self.stdin)
+        if self.stdout_handle != 0:
+            self.handles.add(self.stdout_handle, self.stdout)
+        if self.stderr_handle != 0:
+            self.handles.add(self.stderr_handle, self.stderr)
+
+        # TODO: attempt to extract handles from the dump stream and add them as UnknownObject
+        if self._minidump.handles is not None:
+            by_type: Dict[str, List[minidump.MinidumpHandleDescriptor]] = {}
+            minidump_handle: minidump.MinidumpHandleDescriptor
+            for minidump_handle in self._minidump.handles.handles:
+                type_name = minidump_handle.TypeName
+                if type_name is None:
+                    type_name = "Unknown"
+                if type_name not in by_type:
+                    by_type[type_name] = []
+                by_type[type_name].append(minidump_handle)
+            def default_fn(o):
+                if isinstance(o, bytes):
+                    return o.hex()
+                else:
+                    return o.__dict__
+            for type_name, handles in by_type.items():
+                for minidump_handle in handles:
+                    handle_value = minidump_handle.Handle
+                    handle_data = self.handles.get(handle_value, None)
+                    if handle_data is not None:
+                        self.debug(f"handle already added: {hex(handle_value)} = {self.handles.get(handle_value, None)}")
+                        continue
+
+                    obj: AbstractObject
+                    if type_name == "Unknown":
+                        obj = UnknownObject()
+                    elif type_name == "File":
+                        path = minidump_handle.ObjectName
+                        if path is None:
+                            path = "???"
+                        obj = AbstractFileObject(path)
+                    elif type_name == "Event":
+                        # TODO: parse the ObjectInfos when available
+                        event_type = EVENT_TYPE.SynchronizationEvent
+                        event_signalled = False
+                        obj = EventObject(event_type, event_signalled)
+                    elif type_name == "Key":
+                        key = minidump_handle.ObjectName
+                        if key is None:
+                            key = "???"
+                        obj = RegistryKeyObject(key)
+                    else:
+                        obj = UnsupportedObject(type_name)
+                    self.handles.add(handle_value, obj)
+
     def _setup_emulator(self, thread):
         self._setup_pebteb(thread)
         # TODO: map these using self.memory instead
@@ -640,7 +720,7 @@ class Dumpulator(Architecture):
     def _parse_module_exports(self, module):
         try:
             module_data = self.read(module.baseaddress, module.size)
-        except UcError:
+        except IndexError:
             self.error(f"Failed to read module data")
             return []
         pe = PE(data=module_data, fast_load=True)
@@ -680,7 +760,7 @@ class Dumpulator(Architecture):
                 try:
                     data = self.read(va, size)
                     mapped_data[rva:size] = data
-                except UcError:
+                except IndexError:
                     self.error(f"Failed to read section {name} from module {path}")
             # Load the PE dumped from memory
             pe = PE(data=mapped_data, fast_load=True)
@@ -734,7 +814,12 @@ class Dumpulator(Architecture):
             addr = int(addr)
         self._pages.write(addr, data)
 
-    def call(self, addr, args: List[int] = [], regs: dict = {}, count=0):
+    def call(self, addr, args: List[int] = None, regs: dict = None, count=0):
+        if args is None:
+            args = []
+        if regs is None:
+            regs = {}
+
         if not isinstance(addr, int):
             addr = int(addr)
         # allow passing custom registers
@@ -762,7 +847,7 @@ class Dumpulator(Architecture):
                 start=self._allocate_base,
                 size=self._allocate_size,
                 protect=MemoryProtect.PAGE_EXECUTE_READWRITE,
-                type=MemoryType.MEM_PRIVATE,
+                memory_type=MemoryType.MEM_PRIVATE,
                 info="allocated region"
             )
             self._allocate_ptr = self._allocate_base
@@ -784,10 +869,11 @@ class Dumpulator(Architecture):
         self.exception.handling = True
 
         if self.exception.type == ExceptionType.ContextSwitch:
-            self.info(f"switching context, cip: {hex(self.regs.cip)}")
+            self.info(f"context switch, cip: {hex(self.regs.cip)}")
             # Clear the pending exception
             self.last_exception = self.exception
             self.exception = ExceptionInfo()
+            # NOTE: the context has already been restored using context_restore in the caller
             return self.regs.cip
 
         self.info(f"handling exception...")
@@ -904,7 +990,7 @@ rsp in KiUserExceptionDispatcher:
         self.regs.csp = csp
         return self.KiUserExceptionDispatcher
 
-    def start(self, begin, end=0xffffffffffffffff, count=0):
+    def start(self, begin, end=0xffffffffffffffff, count=0) -> None:
         # Clear exceptions before starting
         self.exception = ExceptionInfo()
         emu_begin = begin
@@ -960,7 +1046,7 @@ rsp in KiUserExceptionDispatcher:
                     traceback.print_exc()
                 break
 
-    def stop(self, exit_code=None):
+    def stop(self, exit_code=None) -> None:
         try:
             self.exit_code = None
             if exit_code is not None:
@@ -986,7 +1072,7 @@ rsp in KiUserExceptionDispatcher:
     def NtCurrentThread(self):
         return 0xFFFFFFFFFFFFFFFE if self._x64 else 0xFFFFFFFE
 
-    def map_module(self, file_data: bytes, file_path: str = "", requested_base: int = 0, resolve_imports = True):
+    def map_module(self, file_data: bytes, file_path: str = "", requested_base: int = 0, resolve_imports=True):
         if not file_path:
             file_path = "<unnamed>"
         print(f"Mapping module {file_path}")
@@ -1010,7 +1096,7 @@ rsp in KiUserExceptionDispatcher:
 
         # https://vtopan.wordpress.com/2019/04/12/patching-resolving-imports-in-a-pe-file-python-pefile/
         # manually resolve imports
-        if hasattr(pe, "DIRECTORY_ENTRY_IMPORT"):
+        if resolve_imports and hasattr(pe, "DIRECTORY_ENTRY_IMPORT"):
             ordinal_flag = 2 ** (bits - 1)
             for iid in pe.DIRECTORY_ENTRY_IMPORT:
                 dll_name = iid.dll.decode("utf-8").lower()
@@ -1112,7 +1198,6 @@ def _hook_code_exception(uc: Uc, address, size, dp: Dumpulator):
         ex.step_count += 1
         if ex.step_count >= ex.tb_icount:
             raise Exception("Stepped past the basic block without reaching exception")
-
     except UcError as err:
         dp.error(f"Exception during unicorn hook, please report this as a bug")
         raise err
@@ -1178,7 +1263,7 @@ def _hook_mem(uc: Uc, access, address, size, value, dp: Dumpulator):
                 assert value == dp.exception.memory_value
 
                 # Delete the code hook
-                uc.hook_del(int(dp.exception.code_hook_h))
+                uc.hook_del(dp.exception.code_hook_h)
                 dp.exception.code_hook_h = None
 
             # At this point we know for sure the context is correct so we can report the exception
@@ -1239,14 +1324,14 @@ def _get_regs(instr, include_write=False):
     return regs
 
 def _hook_code(uc: Uc, address, size, dp: Dumpulator):
+    code = b""
     try:
         code = dp.read(address, min(size, 15))
         instr = next(dp.cs.disasm(code, address, 1))
     except StopIteration:
         instr = None  # Unsupported instruction
-    except UcError:
-        instr = None # Likely invalid memory
-        code = []
+    except IndexError:
+        instr = None  # Likely invalid memory
     address_name = dp.exports.get(address, "")
 
     module = ""
@@ -1272,6 +1357,8 @@ def _hook_code(uc: Uc, address, size, dp: Dumpulator):
             line += instr.op_str
         for reg in _get_regs(instr):
             line += f"|{reg}=0x{dp.regs.__getattr__(reg):x}"
+        if instr.mnemonic in {"syscall", "sysenter"}:
+            line += f"|sequence_id=[{dp.sequence_id}]"
     else:
         line += f"??? (code: {code.hex()}, size: {hex(size)})"
     line += "\n"
@@ -1280,14 +1367,13 @@ def _hook_code(uc: Uc, address, size, dp: Dumpulator):
 def _unicode_string_to_string(dp: Dumpulator, arg: P(UNICODE_STRING)):
     try:
         return arg[0].read_str()
-    except UcError:
-        pass
-    return None
+    except IndexError:
+        return None
 
 def _object_attributes_to_string(dp: Dumpulator, arg: P(OBJECT_ATTRIBUTES)):
     try:
         return arg[0].ObjectName[0].read_str()
-    except UcError:
+    except IndexError:
         pass
     return None
 
@@ -1312,7 +1398,7 @@ def _arg_to_string(dp: Dumpulator, arg):
         if arg.type is OBJECT_ATTRIBUTES:
             tstr = _object_attributes_to_string(dp, arg)
         elif arg.type is UNICODE_STRING:
-            tstr = _unicode_string_to_string(dp, arg)
+            tstr = f"\"{_unicode_string_to_string(dp, arg)}\""
         if tstr is not None:
             str += f" /* {tstr} */"
         return str
@@ -1379,10 +1465,24 @@ def _hook_syscall(uc: Uc, dp: Dumpulator):
                     return dp.regs.r10
                 return dp.args[index]
 
-            dp.info(f"syscall: {name}(")
+            dp.info(f"[{dp.sequence_id}] syscall: {name}(")
             for i in range(0, argcount):
                 argname = argspec.args[1 + i]
                 argtype = argspec.annotations[argname]
+                # Extract the type information from the annotation
+                # Reference: https://github.com/python/cpython/issues/89543
+                # It looks like the python designers did an oopsie, so we're going
+                # the fully-undocumented route.
+                sal = None
+                if argtype.__name__ == "Annotated":
+                    sal, = argtype.__metadata__
+                    argtype = argtype.__origin__
+
+                if sal is None:
+                    sal_pretty = ""
+                else:
+                    sal_pretty = str(sal) + " "
+
                 argvalue = syscall_arg(i)
                 if issubclass(argtype, PVOID):
                     argvalue = argtype(argvalue, dp)
@@ -1399,7 +1499,7 @@ def _hook_syscall(uc: Uc, dp: Dumpulator):
                 if i + 1 == argcount:
                     comma = ""
 
-                dp.info(f"    {_arg_type_string(argvalue)} {argname} = {_arg_to_string(dp, argvalue)}{comma}")
+                dp.info(f"    {sal_pretty}{_arg_type_string(argvalue)} {argname} = {_arg_to_string(dp, argvalue)}{comma}")
             dp.info(")")
             try:
                 status = syscall_impl(dp, *args)
@@ -1410,7 +1510,7 @@ def _hook_syscall(uc: Uc, dp: Dumpulator):
                 else:
                     dp.info(f"status = {status:x}")
                     dp.regs.cax = status
-                    if dp._x64:
+                    if dp.x64:
                         dp.regs.rcx = dp.regs.cip + 2
                         dp.regs.r11 = dp.regs.eflags
             except UcError as err:
@@ -1419,6 +1519,8 @@ def _hook_syscall(uc: Uc, dp: Dumpulator):
                 traceback.print_exc()
                 dp.error(f"Exception thrown during syscall implementation, stopping emulation!")
                 dp.raise_kill(exc)
+            finally:
+                dp.sequence_id += 1
         else:
             dp.error(f"syscall index: {index:x} -> {name} not implemented!")
             dp.raise_kill(NotImplementedError())
@@ -1430,7 +1532,7 @@ def _emulate_unsupported_instruction(dp: Dumpulator, instr: CsInsn):
     if instr.id == X86_INS_RDRAND:
         op: X86Op = instr.operands[0]
         regname = instr.reg_name(op.reg)
-        if dp._x64 and op.size * 8 == 32:
+        if dp.x64 and op.size * 8 == 32:
             regname = "r" + regname[1:]
         print(f"emulated rdrand {regname}:{op.size * 8}, cip = {hex(instr.address)}+{instr.size}")
         dp.regs[regname] = 42  # TODO: PRNG based on dmp hash
@@ -1461,7 +1563,6 @@ def _hook_invalid(uc: Uc, dp: Dumpulator):
             return False  # NOTE: returning True would stop emulation
     except StopIteration:
         pass  # Unsupported instruction
-    except UcError:
+    except IndexError:
         pass  # Invalid memory access (NOTE: this should not be possible actually)
     raise NotImplementedError("TODO: throw invalid instruction exception")
-    return False
