@@ -1,7 +1,5 @@
-import struct
 import ctypes
 import traceback
-from typing import Optional
 
 from .ntenums import *
 from .ntprimitives import *
@@ -21,6 +19,7 @@ STATUS_OBJECT_NAME_NOT_FOUND = 0xC0000034
 STATUS_NOT_FOUND = 0xC0000225
 STATUS_MEMORY_NOT_ALLOCATED = 0xC00000A0
 STATUS_CONFLICTING_ADDRESSES = 0xC0000018
+STATUS_PORT_NOT_SET = 0xC0000353
 
 # Exceptions
 DBG_PRINTEXCEPTION_C = 0x40010006
@@ -83,6 +82,22 @@ IMAGE_SCN_MEM_EXECUTE = 0x20000000
 IMAGE_SCN_MEM_READ = 0x40000000
 IMAGE_SCN_MEM_WRITE = 0x80000000
 
+# Registry value types
+REG_NONE = 0  # No value type
+REG_SZ = 1  # Unicode nul terminated string
+REG_EXPAND_SZ = 2  # Unicode nul terminated string (with environment variable references)
+REG_BINARY = 3  # Free form binary
+REG_DWORD = 4  # 32-bit number
+REG_DWORD_LITTLE_ENDIAN = 4  # 32-bit number (same as REG_DWORD)
+REG_DWORD_BIG_ENDIAN = 5  # 32-bit number
+REG_LINK = 6  # Symbolic Link (unicode)
+REG_MULTI_SZ = 7  # Multiple Unicode strings
+REG_RESOURCE_LIST = 8  # Resource list in the resource map
+REG_FULL_RESOURCE_DESCRIPTOR = 9  # Resource list in the hardware description
+REG_RESOURCE_REQUIREMENTS_LIST = 10  # Resource requirements list
+REG_QWORD = 11  # 64-bit number
+REG_QWORD_LITTLE_ENDIAN = 11  # 64-bit number (same as REG_QWORD)
+
 def round_to_pages(size):
     return (size + 0xFFF) & 0xFFFFFFFFFFFFF000
 
@@ -243,16 +258,14 @@ class CONTEXT(ctypes.Structure):
                     if key.startswith("Seg"):
                         dpname = key[3:].lower()
                     setattr(self, key, regs[dpname])
-                except Exception as x:
+                except Exception:
                     traceback.print_exc()
-                    pass
         if (ContextFlags & CONTEXT_INTEGER) == CONTEXT_INTEGER:
             for key in CONTEXT._integer:
                 try:
                     setattr(self, key, regs[key.lower()])
-                except Exception as x:
+                except Exception:
                     traceback.print_exc()
-                    pass
         if (ContextFlags & CONTEXT_SEGMENTS) == CONTEXT_SEGMENTS:
             for key in CONTEXT._segments:
                 try:
@@ -260,17 +273,15 @@ class CONTEXT(ctypes.Structure):
                     if key.startswith("Seg"):
                         dpname = key[3:].lower()
                     setattr(self, key, regs[dpname])
-                except Exception as x:
+                except Exception:
                     traceback.print_exc()
-                    pass
         if (ContextFlags & CONTEXT_DEBUG_REGISTERS) == CONTEXT_DEBUG_REGISTERS:
             for key in CONTEXT._debug:
                 try:
                     value = regs[key.lower()] if key.startswith("Dr") else 0
                     setattr(self, key, value)
-                except Exception as x:
+                except Exception:
                     traceback.print_exc()
-                    pass
         if (ContextFlags & CONTEXT_MMX_REGISTERS) == CONTEXT_MMX_REGISTERS:
             xmm = self.FltSave.Xmm
             for key in CONTEXT._mmx:
@@ -280,9 +291,8 @@ class CONTEXT(ctypes.Structure):
                 y.Low = x - (x >> 64)
                 try:
                     setattr(xmm, key, y)
-                except Exception as x:
+                except Exception:
                     traceback.print_exc()
-                    pass
 
     def to_regs(self, regs):
         setattr(self, "MxCsr", regs["mxcsr"])
@@ -295,16 +305,14 @@ class CONTEXT(ctypes.Structure):
                     if key.startswith("Seg"):
                         dpname = key[3:].lower()
                     setattr(regs, dpname, getattr(self, key))
-                except Exception as x:
+                except Exception:
                     traceback.print_exc()
-                    pass
         if (ContextFlags & CONTEXT_INTEGER) == CONTEXT_INTEGER:
             for key in CONTEXT._integer:
                 try:
                     setattr(regs, key.lower(), getattr(self, key))
-                except Exception as x:
+                except Exception:
                     traceback.print_exc()
-                    pass
         if (ContextFlags & CONTEXT_SEGMENTS) == CONTEXT_SEGMENTS:
             for key in CONTEXT._segments:
                 try:
@@ -312,17 +320,15 @@ class CONTEXT(ctypes.Structure):
                     if key.startswith("Seg"):
                         dpname = key[3:].lower()
                     setattr(regs, dpname, getattr(self, key))
-                except Exception as x:
+                except Exception:
                     traceback.print_exc()
-                    pass
         if (ContextFlags & CONTEXT_DEBUG_REGISTERS) == CONTEXT_DEBUG_REGISTERS:
             for key in CONTEXT._debug:
                 try:
                     if key.startswith("Dr"):
                         setattr(regs, key.lower(), getattr(self, key))
-                except Exception as x:
+                except Exception:
                     traceback.print_exc()
-                    pass
         if (ContextFlags & CONTEXT_MMX_REGISTERS) == CONTEXT_MMX_REGISTERS:
             # TODO implement
             pass
@@ -566,9 +572,42 @@ def FILE_BASIC_INFORMATION(arch: Architecture):
         ]
     return FILE_BASIC_INFORMATION()
 
+def SECTION_IMAGE_INFORMATION(arch: Architecture):
+    class SECTION_IMAGE_INFORMATION(ctypes.Structure):
+        _alignment_ = arch.alignment()
+        _fields_ = [
+            ("TransferAddress", arch.ptr_type()),
+            ("ZeroBits", ctypes.c_uint32),
+            ("MaximumStackSize", arch.ptr_type()),
+            ("CommittedStackSize", arch.ptr_type()),
+            ("SubSystemType", ctypes.c_uint32),
+            ("SubSystemMinorVersion", ctypes.c_uint16),
+            ("SubSystemMajorVersion", ctypes.c_uint16),
+            ("MajorOperatingSystemVersion", ctypes.c_uint16),
+            ("MinorOperatingSystemVersion", ctypes.c_uint16),
+            ("ImageCharacteristics", ctypes.c_uint16),
+            ("DllCharacteristics", ctypes.c_uint16),
+            ("Machine", ctypes.c_uint16),
+            ("ImageContainsCode", ctypes.c_uint8),
+            ("ImageFlags", ctypes.c_uint8),
+            ("LoaderFlags", ctypes.c_uint32),
+            ("ImageFileSize", ctypes.c_uint32),
+            ("CheckSum", ctypes.c_uint32),
+        ]
+    return SECTION_IMAGE_INFORMATION()
+
 def P(t):
     class P(PVOID):
         def __init__(self, ptr, mem_read):
             super().__init__(ptr, mem_read)
             self.type = t
     return P
+
+class KEY_VALUE_FULL_INFORMATION(ctypes.Structure):
+    _fields_ = [
+        ("TitleIndex", ctypes.c_uint32),
+        ("Type", ctypes.c_uint32),
+        ("DataOffset", ctypes.c_uint32),
+        ("DataLength", ctypes.c_uint32),
+        ("NameLength", ctypes.c_uint32),
+    ]
